; Author: Petroizki
	
; NOTE! To save some calculations i usually compare the distance on square,
;  this means that i don't take the square root from the distance function (sqrt(x*X + y*y)).
	
	AMOUNT_OF_DOTS EQU <1024>		;! must be >0 !!!
	DISTANCE_CURSOR_SQRT EQU <2500.0>; minimum distance to cursor (pow2)
	DISTANCE_MINIMUM_SQRT EQU <1000.0>; minimum distance to other dots (pow2)
	DISTANCE_NEIGHBOUR_SQRT EQU <7000.0> ; distance within dots are threathed as neighbours (pow2)
	SPEED_EXTRA EQU <1.5>			; extra speed that is added, when cursor is near dot
	SPEED_NORMAL EQU <2.5>			; normal speed
	WALL_DISTANCE EQU <20>			; distance to walls the dots try to keep
	WEIGHT_CURSOR EQU <1.7>			; "weight" (strength) of cursor vector, which makes dots avoid cursor
	WEIGHT_WALLS EQU <2.0>			; "weight" of the walls

	MATRIX2x1 struct
		x REAL4 ?
		y REAL4 ?
	MATRIX2x1 ends
	
	SINGLE_DOT struct
		color COLORREF ?			; color of the dot
		plan MATRIX2x1 <?>			; future plans
		position MATRIX2x1 <?>		; position of the dot
		velocity MATRIX2x1 <?>		; velocity of the dot
	SINGLE_DOT ends
	
	CreateDots proto :DWORD
	RunAI proto :REAL4, :REAL4
	UpdateWorld proto :HDC
.data
	fDistanceCursor MATRIX2x1 <DISTANCE_CURSOR_SQRT, DISTANCE_CURSOR_SQRT>
	fDistanceMinimum MATRIX2x1 <DISTANCE_MINIMUM_SQRT, DISTANCE_MINIMUM_SQRT>
	fDistanceNeighbour MATRIX2x1 <DISTANCE_NEIGHBOUR_SQRT, DISTANCE_NEIGHBOUR_SQRT>
	fOneX MATRIX2x1 <WEIGHT_WALLS, 0.0>
	fOneY MATRIX2x1 <0.0, WEIGHT_WALLS>
	fNegativeX dd 80000000h, 0
	fNegativeY dd 0, 80000000h
	fPositiveX dd 7FFFFFFFh, 0FFFFFFFFh
	fPositiveY dd 0FFFFFFFFh, 7FFFFFFFh
	fSpeedExtra MATRIX2x1 <SPEED_EXTRA, SPEED_EXTRA>
	fSpeedNormal MATRIX2x1 <SPEED_NORMAL, SPEED_NORMAL>
	fWeightCursor MATRIX2x1 <WEIGHT_CURSOR, WEIGHT_CURSOR>
.data?
	dwMaxX DWORD ?
	dwMaxY DWORD ?
	dwWorldMaxX DWORD ?
	dwWorldMaxY DWORD ?
	dots SINGLE_DOT AMOUNT_OF_DOTS dup (<?>)
.code
; creates dots with nrandomom position and color
CreateDots proc uses ebx edi WorldRect:DWORD
	LOCAL temp:DWORD
	
	mov edx, dword ptr [WorldRect]		; get world rect
	
	; here we make the wall minimum distances
	mov eax, dword ptr [edx].RECT.right		; get right wall pos
	mov dword ptr [dwWorldMaxX], eax
	sub eax, WALL_DISTANCE						; subtract the minimum distance to wall
	mov dword ptr [dwMaxX], eax			 	; save the maximum x pos
	
	mov eax, dword ptr [edx].RECT.bottom		; get bottom wall pos
	mov dword ptr [dwWorldMaxY], eax
	sub eax, WALL_DISTANCE						; subtract the minimum distance to wall
	mov dword ptr [dwMaxY], eax				; save the maximum y pos
	
	mov edi, OFFSET dots
	mov ebx, AMOUNT_OF_DOTS
	
	; creates nrandomom colored dot to nrandomom position
	.REPEAT
		mov dword ptr [edi].SINGLE_DOT.velocity.x, 40000000h
		mov dword ptr [edi].SINGLE_DOT.velocity.y, 40000000h
		mov dword ptr [edi].SINGLE_DOT.plan.x, 40000000h
		mov dword ptr [edi].SINGLE_DOT.plan.y, 40000000h
		
		invoke nrandom, 0FFFFFFh
		mov dword ptr [edi].SINGLE_DOT.color, eax
		
		invoke nrandom, SCREEN_WIDTH
		
		push eax
		fild dword ptr [esp]
		fstp dword ptr [edi].SINGLE_DOT.position.x
		mov dword ptr [esp], SCREEN_HEIGHT
		call nrandom
		
		mov dword ptr [temp], eax
		fild dword ptr [temp]
		fstp dword ptr [edi].SINGLE_DOT.position.y
		
		add edi, SIZEOF SINGLE_DOT
		dec ebx
	.UNTIL zero?
	
	ret
CreateDots endp

; runs the A.I. for all dots, this will set every dots velocity
RunAI proc uses ebx edi esi, fMousePosX:REAL4, fMousePosY:REAL4
	LOCAL pos:MATRIX2x1
	; forces that affect the dots:
	; 1. Alignment
	;     Dot will align the velocity vector with local flock (neighborhood dots).
	; 2. Separation
	;     Dot will keep a minimum distance from local dots.
	; 3. Cohesion
	;     Dot will try to fly the center position of local flock.
	; 4. Evasion
	;     Dot will try to evade mouse cursor.
	;     Dot will evade world edges.
	;      there are two world edge checks, other one creates an vector
	;      from the wall, and hints the dot to go another way,
	;      the second check just forces the dot to stay within the window
	femms
	
	mov ecx, AMOUNT_OF_DOTS
	mov edx, OFFSET dots
	
	.REPEAT
		xor ebx, ebx
		
		; these will be here trough the loop
		movq mm6, qword ptr [edx].SINGLE_DOT.plan
		movq mm7, qword ptr [edx].SINGLE_DOT.position
		
		lea edi, [ecx - 1]
		
		test edi, edi
		jz @end_of_neighbours
		
		lea esi, [edx + SIZEOF SINGLE_DOT]
@neighbour_check:
			; creates an vector to the current nearby dot
			movq mm0, qword ptr [esi].SINGLE_DOT.position ; pos y | pos x
			pfsub mm0, mm7							; vector y | vector x
			movq mm1, mm0
			
			; makes an square length of the vector
			pfmul mm0, mm0							; y*y | x*x
			pfacc mm0, mm0							; x*x + y*y | x*x + y*y
			
			; check if this dot is considered as "neighbour"
			movq mm2, mm0
			pfcmpgt mm0, qword ptr [fDistanceNeighbour]
			movd eax, mm0
			test eax, eax
			jnz @not_neighbour
			
			pfrsqrt mm0, mm2						; 1/sqrt(x*x + y*y) | 1/sqrt(x*x + y*y)
			pfmul mm1, mm0							; y/sqrt(x*x + y*y) | x/sqrt(x*x + y*y)
			
			; The dot is aligned the same way as dots nearby
			movq mm0, qword ptr [esi].SINGLE_DOT.plan
			pfadd mm0, qword ptr [edx].SINGLE_DOT.velocity
			
			; add the nearby dots velocity to current dots plan
			pfadd mm6, qword ptr [esi].SINGLE_DOT.velocity
			
			; dot will either wants to either pursue or avoid, the dot
			; nearby, depending how close it is
			; If the other dot is too close, we must go away
			pfcmpgt mm2, qword ptr [fDistanceMinimum]
			movd eax, mm2
			test eax, eax
			jnz @not_near
			
			; update plan
			; add an vector from the other dot, because we don't want to go any closer
			pfsub mm6, mm1
			pfadd mm0, mm1
			jmp @F
@not_near:	; update plan
			; add an vector to the other dot, because we are not close enough...
			pfadd mm6, mm1
			pfsub mm0, mm1
@@:			movq qword ptr [esi].SINGLE_DOT.plan, mm0
@not_neighbour:
			add esi, SIZEOF SINGLE_DOT
			dec edi
		jnz @neighbour_check
@end_of_neighbours:
		
		; The plan is normalized here
		; so the "weights" of different forces are equal
		movq mm0, mm6
		pfmul mm0, mm0
		pfacc mm0, mm0
		pfrsqrt mm0, mm0
		pfmul mm6, mm0
		
		; - Dot will evade mouse cursor
		;    we will just create vector to opposite direction to cursor,
		;    and add it to the final plan
		movq mm0, mm7
		pfsub mm0, qword ptr [fMousePosX]
		movq mm1, mm0
		pfmul mm0, mm0
		pfacc mm0, mm0
		movq mm2, mm0
		pfcmpgt mm0, qword ptr [fDistanceCursor]
		movd eax, mm0
		test eax, eax
		jnz @no_mouse_cursor
		
		pfrsqrt mm0, mm2
		pfmul mm1, mm0
		
		pfmul mm1, qword ptr [fWeightCursor]		; add the "weight" of mouse cursor
		pfadd mm6, mm1
		or ebx, -1									; set mouse cursor nearby
@no_mouse_cursor:
		
		; - The old velocity will affect the plan
		pfadd mm6, qword ptr [edx].SINGLE_DOT.velocity
		
		; - Dot will try to evade walls
		;    Here we create an normal vector from wall, which is
		;    too near, and add it to the plan
		movq mm0, mm7
		pf2id mm0, mm0
		movq qword ptr [pos], mm0
		
		mov eax, dword ptr [pos].MATRIX2x1.x
		
		cmp eax, WALL_DISTANCE
		ja @check_right_wall
		
		pfadd mm6, qword ptr [fOneX]
		
		cmp eax, 0
		ja @F
		
		pand mm6, qword ptr [fPositiveX]
@@:		jmp @check_top_wall
@check_right_wall:
		cmp eax, dword ptr [dwMaxX]
		jb @check_top_wall
		
		pfsub mm6, qword ptr [fOneX]
		
		cmp eax, dword ptr [dwWorldMaxX]
		jb @check_top_wall
		
		por mm6, qword ptr [fNegativeX]
@check_top_wall:
		mov eax, dword ptr [pos].MATRIX2x1.y
		
		cmp eax, WALL_DISTANCE
		ja @check_bottom_wall
		
		pfadd mm6, qword ptr [fOneY]
		
		cmp eax, 0
		ja @F
		
		pand mm6, qword ptr [fPositiveY]
@@:		jmp @end_wall_check
@check_bottom_wall:
		cmp eax, dword ptr [dwMaxY]
		jb @end_wall_check
		
		pfsub mm6, qword ptr [fOneY]
		
		cmp eax, dword ptr [dwWorldMaxY]
		jb @end_wall_check
		
		por mm6, qword ptr [fNegativeY]
@end_wall_check:
		; - The plan will be made as the current velocity,
		;    after it has been normalized and scaled to desired speed
		movq mm0, mm6
		pfmul mm0, mm0
		pfacc mm0, mm0
		pfrsqrt mm0, mm0
		pfmul mm0, mm6
		
		; add speed
		pfmul mm0, qword ptr [fSpeedNormal]
		test ebx, ebx
		jz @F
		pfmul mm0, qword ptr [fSpeedExtra]			; add extra speed if mouse cursor is nearby
@@:
		movq qword ptr [edx].SINGLE_DOT.velocity, mm0
		pxor mm0, mm0
		movq qword ptr [edx].SINGLE_DOT.plan, mm0
		
		add edx, SIZEOF SINGLE_DOT
		dec ecx
	.UNTIL zero?
	
	femms
	
	ret
RunAI endp

; updates dots position, and draws them
UpdateWorld proc uses ebx edi esi, hDC:HDC
	mov ebx, AMOUNT_OF_DOTS
	mov esi, [hDC]
	mov edi, OFFSET dots
	
	femms
	
	.REPEAT
		push dword ptr [edi].SINGLE_DOT.color		; 3
		sub esp, 8									; make room for the two other params (x and y)
		
		movq mm0, qword ptr [edi].SINGLE_DOT.position
		movq mm1, qword ptr [edi].SINGLE_DOT.velocity
		
		pfadd mm0, mm1								; update position by adding an velocity vector
		movq qword ptr [edi].SINGLE_DOT.position, mm0; save updated position
		pf2id mm0, mm0
		movq qword ptr [esp], mm0
		
		push esi									; 0 - DC handle
		femms
		call SetPixel								; draw dot
		
		add edi, SIZEOF SINGLE_DOT					; goto next dot struct
		dec ebx										; decrease counter
	.UNTIL zero?
	
	ret
UpdateWorld endp